# Artifact for "Foundations for Cryptographic Reductions in CCSA Logics"

Authors: David Baelde, Adrien Koutsos and Justine Sauvage

Institutions: Univ Rennes, CNRS, IRISA and Inria Paris

Conference: CCS 2024

Pre-print available at: https://hal.science/hal-04511718

## Abstract of the paper

The Computationally Complete Symbolic Attacker (CCSA) approach to security
protocol verification relies on probabilistic logics to reason about the
interaction traces between a protocol and an arbitrary adversary. The proof
assistant Squirrel implements one such logic. CCSA logics come with
cryptographic axioms whose soundness derives from the security of standard
cryptographic games, e.g. PRF, EUF, IND-CCA. Unfortunately, these axioms are
complex to design and implement; so far, these tasks are manual, ad hoc and
error-prone. We solve these issues by providing a formal and systematic method
for deriving axioms from cryptographic games. Our method relies on
synthesizing an adversary against some cryptographic game, through the notion
of bi-deduction. Concretely, we define a rich notion of bi-deduction, justify
how to use it to derive cryptographic axioms, provide a proof system for
bi-deduction, and an automatic proof-search method which we implemented in
Squirrel.

## Artifact contents

The `squirrel-prover` subdirectory contains the source code for the Squirrel
prover, extended with the "crypto" tactic introduced in the paper. The purpose
of this tactic is to replace legacy Squirrel tactics implementing cryptographic
axioms. The "crypto" tactic works by bi-deducing equivalences from
cryptographic games specified by the user.

Detailed instructions on how to build and use Squirrel are provided in
`squirrel-prover/README.md`.

The case studies presented in the paper are provided in the
subdirectory `squirrel-prover/examples/crypto`. These files are all accepted
by our modified version of Squirrel, and contain definitions of cryptographic
games as well as uses of the "crypto" tactic using these games, as part of
larger proof scripts establishing various security properties.

The `case-studies-html` subdirectory contains HTML exports of Squirrel's
run on each of the case studies, which can be replayed in the browser.
This allows to replay, without building or installing anything, each step
of the presented developments.
